from __future__ import annotations

from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING

import pytest

from infrahub_sdk.schema.main import AttributeKind, NodeSchema, RelationshipKind, SchemaRoot
from infrahub_sdk.schema.main import AttributeSchema as Attr
from infrahub_sdk.schema.main import RelationshipSchema as Rel

if TYPE_CHECKING:
    from infrahub_sdk import InfrahubClient
    from infrahub_sdk.node import InfrahubNode

NAMESPACE = "Testing"

TESTING_MANUFACTURER = f"{NAMESPACE}Manufacturer"
TESTING_PERSON = f"{NAMESPACE}Person"
TESTING_CAR = f"{NAMESPACE}Car"
BUILTIN_TAG = "BuiltinTag"


@dataclass
class TestingPersonData:
    name: str
    kind: str = TESTING_PERSON


@dataclass
class TestingManufacturerData:
    name: str
    kind: str = TESTING_MANUFACTURER


@dataclass
class TestingCarData:
    name: str
    color: str | None = None
    kind: str = TESTING_CAR


class SchemaCarPerson:
    @pytest.fixture(scope="class")
    def schema_person_base(self) -> NodeSchema:
        return NodeSchema(
            name="Person",
            namespace=NAMESPACE,
            include_in_menu=True,
            label="Person",
            human_friendly_id=["name__value"],
            attributes=[
                Attr(name="name", kind=AttributeKind.TEXT, unique=True),
                Attr(name="description", kind=AttributeKind.TEXT, optional=True),
                Attr(name="height", kind=AttributeKind.NUMBER, optional=True),
                Attr(name="age", kind=AttributeKind.NUMBER, optional=True),
            ],
            relationships=[
                Rel(name="cars", kind=RelationshipKind.GENERIC, optional=True, peer=TESTING_CAR, cardinality="many")
            ],
        )

    @pytest.fixture(scope="class")
    def schema_car_base(self) -> NodeSchema:
        return NodeSchema(
            name="Car",
            namespace=NAMESPACE,
            include_in_menu=True,
            default_filter="name__value",
            human_friendly_id=["owner__name__value", "name__value"],
            label="Car",
            attributes=[
                Attr(name="name", kind=AttributeKind.TEXT),
                Attr(name="description", kind=AttributeKind.TEXT, optional=True),
                Attr(name="color", kind=AttributeKind.TEXT),
            ],
            relationships=[
                Rel(
                    name="owner",
                    kind=RelationshipKind.ATTRIBUTE,
                    optional=False,
                    peer=TESTING_PERSON,
                    cardinality="one",
                ),
                Rel(
                    name="manufacturer",
                    kind=RelationshipKind.ATTRIBUTE,
                    optional=False,
                    peer=TESTING_MANUFACTURER,
                    cardinality="one",
                    identifier="car__manufacturer",
                ),
                Rel(
                    name="tags",
                    optional=True,
                    peer=BUILTIN_TAG,
                    cardinality="many",
                ),
            ],
        )

    @pytest.fixture(scope="class")
    def schema_manufacturer_base(self) -> NodeSchema:
        return NodeSchema(
            name="Manufacturer",
            namespace=NAMESPACE,
            include_in_menu=True,
            label="Manufacturer",
            human_friendly_id=["name__value"],
            attributes=[
                Attr(name="name", kind=AttributeKind.TEXT),
                Attr(name="description", kind=AttributeKind.TEXT, optional=True),
            ],
            relationships=[
                Rel(
                    name="cars",
                    kind=RelationshipKind.GENERIC,
                    optional=True,
                    peer=TESTING_CAR,
                    cardinality="many",
                    identifier="car__manufacturer",
                ),
                Rel(
                    name="customers",
                    kind=RelationshipKind.GENERIC,
                    optional=True,
                    peer=TESTING_PERSON,
                    cardinality="many",
                    identifier="person__manufacturer",
                ),
            ],
        )

    @pytest.fixture(scope="class")
    def schema_base(
        self,
        schema_car_base: NodeSchema,
        schema_person_base: NodeSchema,
        schema_manufacturer_base: NodeSchema,
    ) -> SchemaRoot:
        return SchemaRoot(version="1.0", nodes=[schema_car_base, schema_person_base, schema_manufacturer_base])

    @pytest.fixture(scope="class")
    async def person_joe_data(self) -> TestingPersonData:
        return TestingPersonData(name="Joe Doe")

    @pytest.fixture(scope="class")
    async def person_jane_data(self) -> TestingPersonData:
        return TestingPersonData(name="Jane Doe")

    @pytest.fixture(scope="class")
    async def manufacturer_vw_data(self) -> TestingManufacturerData:
        return TestingManufacturerData(name="Volkswagen")

    @pytest.fixture(scope="class")
    async def manufacturer_renault_data(self) -> TestingManufacturerData:
        return TestingManufacturerData(name="Renault")

    @pytest.fixture(scope="class")
    async def manufacturer_mercedes_data(self) -> TestingManufacturerData:
        return TestingManufacturerData(name="Mercedes")

    @pytest.fixture(scope="class")
    async def car_golf_data(self) -> TestingCarData:
        return TestingCarData(name="Golf", color="Black")

    @pytest.fixture(scope="class")
    async def person_joe(self, client: InfrahubClient, person_joe_data: TestingPersonData) -> InfrahubNode:
        obj = await client.create(**asdict(person_joe_data))
        await obj.save()
        return obj

    @pytest.fixture(scope="class")
    async def person_jane(self, client: InfrahubClient, person_jane_data: TestingPersonData) -> InfrahubNode:
        obj = await client.create(**asdict(person_jane_data))
        await obj.save()
        return obj

    @pytest.fixture(scope="class")
    async def manufacturer_mercedes(
        self, client: InfrahubClient, manufacturer_mercedes_data: TestingManufacturerData
    ) -> InfrahubNode:
        obj = await client.create(**asdict(manufacturer_mercedes_data))
        await obj.save()
        return obj

    @pytest.fixture(scope="class")
    async def car_golf(
        self,
        client: InfrahubClient,
        car_golf_data: TestingCarData,
        manufacturer_mercedes: InfrahubNode,
        person_joe: InfrahubNode,
    ) -> InfrahubNode:
        obj = await client.create(**asdict(car_golf_data), manufacturer=manufacturer_mercedes, owner=person_joe)
        await obj.save()
        return obj

    @pytest.fixture(scope="class")
    async def tag_blue(self, client: InfrahubClient) -> InfrahubNode:
        obj = await client.create(kind=BUILTIN_TAG, name="Blue")
        await obj.save()
        return obj

    @pytest.fixture(scope="class")
    async def tag_red(self, client: InfrahubClient) -> InfrahubNode:
        obj = await client.create(kind=BUILTIN_TAG, name="Red")
        await obj.save()
        return obj

    @pytest.fixture(scope="class")
    async def tag_green(self, client: InfrahubClient) -> InfrahubNode:
        obj = await client.create(kind=BUILTIN_TAG, name="Green")
        await obj.save()
        return obj

    async def create_persons(self, client: InfrahubClient, branch: str) -> list[InfrahubNode]:
        john = await client.create(kind=TESTING_PERSON, name="John Doe", branch=branch)
        await john.save()

        jane = await client.create(kind=TESTING_PERSON, name="Jane Doe", branch=branch)
        await jane.save()

        return [john, jane]

    async def create_manufacturers(self, client: InfrahubClient, branch: str) -> list[InfrahubNode]:
        obj1 = await client.create(kind=TESTING_MANUFACTURER, name="Volkswagen", branch=branch)
        await obj1.save()

        obj2 = await client.create(kind=TESTING_MANUFACTURER, name="Renault", branch=branch)
        await obj2.save()

        obj3 = await client.create(kind=TESTING_MANUFACTURER, name="Mercedes", branch=branch)
        await obj3.save()

        return [obj1, obj2, obj3]

    async def create_initial_data(self, client: InfrahubClient, branch: str) -> dict[str, list[InfrahubNode]]:
        persons = await self.create_persons(client=client, branch=branch)
        manufacturers = await self.create_manufacturers(client=client, branch=branch)

        car10 = await client.create(
            kind=TESTING_CAR, name="Golf", color="Black", manufacturer=manufacturers[0].id, owner=persons[0].id
        )
        await car10.save()

        car20 = await client.create(
            kind=TESTING_CAR, name="Megane", color="Red", manufacturer=manufacturers[1].id, owner=persons[1].id
        )
        await car20.save()

        return {TESTING_PERSON: persons, TESTING_CAR: [car10, car20], TESTING_MANUFACTURER: manufacturers}



from infrahub.core.constants import (
    NAMESPACE_REGEX,
)

from ...attribute_schema import AttributeSchema as Attr
from ...generic_schema import GenericSchema
from ...node_schema import NodeSchema

generic_menu_item = GenericSchema(
    name="Menu",
    namespace="Core",
    include_in_menu=False,
    description="Element of the Menu",
    label="Menu",
    hierarchical=True,
    human_friendly_id=["namespace__value", "name__value"],
    display_labels=["label__value"],
    generate_profile=False,
    attributes=[
        Attr(name="namespace", kind="Text", regex=NAMESPACE_REGEX, order_weight=1000),
        Attr(name="name", kind="Text", order_weight=1000),
        Attr(name="label", kind="Text", optional=True, order_weight=2000),
        Attr(name="kind", kind="Text", optional=True, order_weight=2500),
        Attr(name="path", kind="Text", optional=True, order_weight=2500),
        Attr(name="description", kind="Text", optional=True, order_weight=3000),
        Attr(name="icon", kind="Text", optional=True, order_weight=4000),
        Attr(name="protected", kind="Boolean", default_value=False, read_only=True, order_weight=5000),
        Attr(name="order_weight", kind="Number", default_value=2000, order_weight=6000),
        Attr(name="required_permissions", kind="List", optional=True, order_weight=7000),
        Attr(
            name="section",
            kind="Text",
            enum=["object", "internal"],
            default_value="object",
            order_weight=8000,
        ),
    ],
)

menu_item = NodeSchema(
    name="MenuItem",
    namespace="Core",
    include_in_menu=False,
    description="Menu Item",
    label="Menu Item",
    inherit_from=["CoreMenu"],
    generate_profile=False,
)



from typing import TYPE_CHECKING

from prefect import get_run_logger, task
from prefect.automations import AutomationCore
from prefect.cache_policies import NONE
from prefect.client.orchestration import PrefectClient
from prefect.client.schemas.filters import DeploymentFilter, DeploymentFilterName
from prefect.events.schemas.automations import Automation

from infrahub.trigger.models import TriggerDefinition

from .models import TriggerSetupReport, TriggerType

if TYPE_CHECKING:
    from uuid import UUID


def compare_automations(target: AutomationCore, existing: Automation) -> bool:
    """Compare an AutomationCore with an existing Automation object to identify if they are identical or not

    Return True if the target is identical to the existing automation
    """

    target_dump = target.model_dump(exclude_defaults=True, exclude_none=True)
    existing_dump = existing.model_dump(exclude_defaults=True, exclude_none=True, exclude={"id"})

    return target_dump == existing_dump


@task(name="trigger-setup", task_run_name="Setup triggers", cache_policy=NONE)  # type: ignore[arg-type]
async def setup_triggers(
    client: PrefectClient,
    triggers: list[TriggerDefinition],
    trigger_type: TriggerType | None = None,
    force_update: bool = False,
) -> TriggerSetupReport:
    log = get_run_logger()

    report = TriggerSetupReport()

    if trigger_type:
        log.info(f"Setting up triggers of type {trigger_type.value}")
    else:
        log.info("Setting up all triggers")

    # -------------------------------------------------------------
    # Retrieve existing Deployments and Automation from the server
    # -------------------------------------------------------------
    deployment_names = list({name for trigger in triggers for name in trigger.get_deployment_names()})
    deployments = {
        item.name: item
        for item in await client.read_deployments(
            deployment_filter=DeploymentFilter(name=DeploymentFilterName(any_=deployment_names))
        )
    }
    deployments_mapping: dict[str, UUID] = {name: item.id for name, item in deployments.items()}

    # If a trigger type is provided, narrow down the list of existing triggers to know which one to delete
    existing_automations: dict[str, Automation] = {}
    if trigger_type:
        existing_automations = {
            item.name: item for item in await client.read_automations() if item.name.startswith(trigger_type.value)
        }
    else:
        existing_automations = {item.name: item for item in await client.read_automations()}

    trigger_names = [trigger.generate_name() for trigger in triggers]
    automation_names = list(existing_automations.keys())

    log.debug(f"{len(automation_names)} existing triggers ({automation_names})")
    log.debug(f"{len(trigger_names)} triggers to configure ({trigger_names})")

    to_delete = set(automation_names) - set(trigger_names)
    log.debug(f"{len(to_delete)} triggers to delete ({to_delete})")

    # -------------------------------------------------------------
    # Create or Update all triggers
    # -------------------------------------------------------------
    for trigger in triggers:
        automation = AutomationCore(
            name=trigger.generate_name(),
            description=trigger.get_description(),
            enabled=True,
            trigger=trigger.trigger.get_prefect(),
            actions=[action.get_prefect(mapping=deployments_mapping) for action in trigger.actions],
        )

        existing_automation = existing_automations.get(trigger.generate_name(), None)

        if existing_automation:
            if force_update or not compare_automations(target=automation, existing=existing_automation):
                await client.update_automation(automation_id=existing_automation.id, automation=automation)
                log.info(f"{trigger.generate_name()} Updated")
                report.updated.append(trigger)
            else:
                report.unchanged.append(trigger)
        else:
            await client.create_automation(automation=automation)
            log.info(f"{trigger.generate_name()} Created")
            report.created.append(trigger)

    # -------------------------------------------------------------
    # Delete Triggers that shouldn't be there
    # -------------------------------------------------------------
    for item_to_delete in to_delete:
        existing_automation = existing_automations.get(item_to_delete)

        if not existing_automation:
            continue

        report.deleted.append(existing_automation)
        await client.delete_automation(automation_id=existing_automation.id)
        log.info(f"{item_to_delete} Deleted")

    return report



from infrahub.core.relationship.constraints.count import RelationshipCountConstraint
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class RelationshipCountConstraintDependency(DependencyBuilder[RelationshipCountConstraint]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> RelationshipCountConstraint:
        return RelationshipCountConstraint(db=context.db, branch=context.branch)
